-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt
local M = {}

local max = math.max
local min = math.min
local abs = math.abs

local constants = { rpmToAV = 0.104719755, avToRPM = 9.549296596425384 }

local gearbox = nil
local engine = nil
local clutch = nil
local hybrid = nil

local sharedFunctions = nil
local gearboxAvailableLogic = nil
local gearboxLogic = nil

M.didInitSecondStage = false
M.gearboxHandling = nil
M.timer = nil
M.timerConstants = nil
M.inputValues = nil
M.shiftPreventionData = nil
M.shiftBehavior = nil
M.smoothedValues = nil
M.needGearboxModeChange = false

M.currentGearIndex = 0
M.throttle = 0
M.brake = 0
M.throttleInput = 0
M.isArcadeSwitched = false

M.smoothedAvgAVInput = 0
M.rpm = 0
M.idleRPM = 0
M.maxRPM = 0

M.engineThrottle = 0
M.engineLoad = 0
M.engineTorque = 0
M.gearboxTorque = 0

M.ignition = true
M.isEngineRunning = 0

M.oilTemp = 0
M.waterTemp = 0
M.checkEngine = false

M.energyStorages = {}

local automaticHandling = {
	availableModes = { "P", "R", "N", "D", "L" },
	hShifterModeLookup = { [-1] = "R", [0] = "N", "P", "D", "L" },
	ecvtGearIndexLookup = { P = -2, R = -1, N = 0, D = 1, L = 2 },
	availableModeLookup = {},
	existingModeLookup = {},
	modeIndexLookup = {},
	modes = {},
	mode = nil,
	modeIndex = 0,
	maxAllowedGearIndex = 0,
	minAllowedGearIndex = 0,
}

local function getGearName()
	return automaticHandling.mode
end

local function getGearPosition()
	return (automaticHandling.modeIndex - 1) / (#automaticHandling.modes - 1)
end

local function applyGearboxModeRestrictions()
	local manualModeIndex
	if string.sub(automaticHandling.mode, 1, 1) == "M" then
		manualModeIndex = string.sub(automaticHandling.mode, 2)
	end
	local maxGearIndex = gearbox.maxGearIndex
	local minGearIndex = gearbox.minGearIndex
	if automaticHandling.mode == "1" then
		maxGearIndex = 1
		minGearIndex = 1
	elseif automaticHandling.mode == "2" then
		maxGearIndex = 2
		minGearIndex = 1
	elseif manualModeIndex then
		maxGearIndex = manualModeIndex
		minGearIndex = manualModeIndex
	end

	automaticHandling.maxGearIndex = maxGearIndex
	automaticHandling.minGearIndex = minGearIndex
end

local function validateGearboxBehavior(behavior)
	if behavior ~= "realistic" then
		if controller.mainController then
			gui.message({ txt = "Arcade mode is not supported by the eCVT gearbox" }, 2, "vehicle.shiftermode.notsupported")
			controller.mainController.setGearboxMode("realistic")
		else
			-- Got called before controller.mainController was set; we will need to override the mode later
			M.needGearboxModeChange = true
		end

		return false
	end

	return true
end

local function gearboxBehaviorChanged(behavior)
	if not validateGearboxBehavior(behavior) then
		return
	end

	gearboxLogic = gearboxAvailableLogic[behavior]
	M.updateGearboxGFX = gearboxLogic.inGear
	M.shiftUp = gearboxLogic.shiftUp
	M.shiftDown = gearboxLogic.shiftDown
	M.shiftToGearIndex = gearboxLogic.shiftToGearIndex
end

local function applyGearboxMode()
	local autoIndex = automaticHandling.modeIndexLookup[automaticHandling.mode]
	if autoIndex then
		automaticHandling.modeIndex = min(max(autoIndex, 1), #automaticHandling.modes)
		automaticHandling.mode = automaticHandling.modes[automaticHandling.modeIndex]
	end

	if automaticHandling.mode == "P" then
		gearbox:setMode("park")
	elseif automaticHandling.mode == "N" then
		gearbox:setMode("neutral")
	elseif automaticHandling.mode == "R" then
		gearbox:setMode("reverse")
	else
		gearbox:setMode("drive")
	end
end

local function shiftUp()
	if automaticHandling.mode == "N" then
		M.timer.gearChangeDelayTimer = M.timerConstants.gearChangeDelay
	end

	automaticHandling.modeIndex = min(automaticHandling.modeIndex + 1, #automaticHandling.modes)
	automaticHandling.mode = automaticHandling.modes[automaticHandling.modeIndex]

	applyGearboxMode()
	applyGearboxModeRestrictions()
end

local function shiftDown()
	if automaticHandling.mode == "N" then
		M.timer.gearChangeDelayTimer = M.timerConstants.gearChangeDelay
	end

	automaticHandling.modeIndex = max(automaticHandling.modeIndex - 1, 1)
	automaticHandling.mode = automaticHandling.modes[automaticHandling.modeIndex]

	applyGearboxMode()
	applyGearboxModeRestrictions()
end

local function shiftToGearIndex(index)
	local desiredMode = automaticHandling.hShifterModeLookup[index]
	if not desiredMode or not automaticHandling.existingModeLookup[desiredMode] then
		if desiredMode and not automaticHandling.existingModeLookup[desiredMode] then
			gui.message({ txt = "vehicle.drivetrain.cannotShiftAuto", context = { mode = desiredMode } }, 2, "vehicle.shiftLogic.cannotShift")
		end
		desiredMode = "N"
	end
	automaticHandling.mode = desiredMode

	applyGearboxMode()
	applyGearboxModeRestrictions()
end

local function updateExposedData()
	M.rpm = engine and (engine.outputAV1 * constants.avToRPM) or 0
	M.smoothedAvgAVInput = sharedFunctions.updateAvgAVSingleDevice("gearbox")
	M.waterTemp = (engine and engine.thermals) and (engine.thermals.coolantTemperature and engine.thermals.coolantTemperature or engine.thermals.oilTemperature) or 0
	M.oilTemp = (engine and engine.thermals) and engine.thermals.oilTemperature or 0
	M.checkEngine = engine and engine.isDisabled or false
	M.ignition = hybrid and hybrid.powerOn or false
	M.engineThrottle = (engine and engine.isDisabled) and 0 or M.throttle
	M.engineLoad = engine and (engine.isDisabled and 0 or engine.instantEngineLoad) or 0
	M.running = hybrid and hybrid.powerOn or false
	M.engineTorque = engine and engine.combustionTorque or 0
	M.flywheelTorque = engine and engine.outputTorque1 or 0
	M.gearboxTorque = gearbox and gearbox.outputTorque or 0
	M.isEngineRunning = hybrid and (hybrid.powerSwitch) and 1 or 0
end

local function updateInputClutch(dt)
	if gearbox and clutch then
		-- Handle gearbox input engaging (clutch is always engaged)

		electrics.values[clutch.electricsClutchRatioName] = 1
	end
end

local function initSecondStage()
	hybrid = controller.getController("hybridPowertrain")
	updateExposedData()

	if M.needGearboxModeChange then -- We will definitely have a controller.mainController now
		gui.message({ txt = "Arcade mode is not supported by the eCVT gearbox" }, 2, "vehicle.shiftermode.notsupported")
		controller.mainController.setGearboxMode("realistic")
	end
end

local function updateInGear(dt)
	if not M.didInitSecondStage then
		initSecondStage() -- why do they not allow shiftLogic to have an actual initSecondStage?!
		M.didInitSecondStage = true
	end

	local gearIndex = automaticHandling.ecvtGearIndexLookup[automaticHandling.mode]

	M.currentGearIndex = (automaticHandling.mode == "N" or automaticHandling.mode == "P") and 0 or gearIndex
	M.throttle = M.inputValues.throttle
	M.brake = M.inputValues.brake
	M.isArcadeSwitched = false

	electrics.values.lockupClutchRatio = 0

	updateExposedData()
	updateInputClutch(dt)
end

local function sendTorqueData()
	if engine then
		engine:sendTorqueData()
	end
end

local function init(jbeamData, sharedFunctionTable)
	sharedFunctions = sharedFunctionTable
	engine = powertrain.getDevice("mainEngine")
	gearbox = powertrain.getDevice("gearbox")
	clutch = powertrain.getDevice("clutch")

	M.currentGearIndex = 0
	M.throttle = 0
	M.brake = 0
	M.clutchRatio = 0

	gearboxAvailableLogic = { realistic = { inGear = updateInGear, shiftUp = shiftUp, shiftDown = shiftDown, shiftToGearIndex = shiftToGearIndex } }
	automaticHandling.availableModeLookup = {}

	for _, v in pairs(automaticHandling.availableModes) do
		automaticHandling.availableModeLookup[v] = true
	end

	automaticHandling.modes = {}
	automaticHandling.modeIndexLookup = {}

	local modes = jbeamData.automaticModes or "PRNDL"
	local modeCount = #modes
	local modeOffset = 0

	for i = 1, modeCount do
		local mode = modes:sub(i, i)
		if automaticHandling.availableModeLookup[mode] then
			automaticHandling.modes[i + modeOffset] = mode
			automaticHandling.modeIndexLookup[mode] = i + modeOffset
			automaticHandling.existingModeLookup[mode] = true
		else
			print("unknown auto mode: " .. mode)
		end
	end

	local defaultMode = jbeamData.defaultAutomaticMode or "N"

	automaticHandling.modeIndex = string.find(modes, defaultMode)
	automaticHandling.mode = automaticHandling.modes[automaticHandling.modeIndex]
	automaticHandling.maxGearIndex = gearbox.maxGearIndex
	automaticHandling.minGearIndex = gearbox.minGearIndex

	M.maxRPM = engine and engine.maxRPM or 0
	M.idleRPM = engine and engine.idleRPM or 0
	M.energyStorages = sharedFunctions.getEnergyStorages({ engine })

	applyGearboxMode()
end

M.init = init

M.gearboxBehaviorChanged = gearboxBehaviorChanged
M.shiftUp = shiftUp
M.shiftDown = shiftDown
M.shiftToGearIndex = shiftToGearIndex
M.updateGearboxGFX = nop
M.getGearName = getGearName
M.getGearPosition = getGearPosition
M.sendTorqueData = sendTorqueData

return M
